import re
import sys

from pydantic import Field, validate_arguments

__all__ = ["build_prefixed_regex_pattern", "build_prefixed_regex_pattern_with_value"]

# Package only support >= 3.8
if sys.version_info[:2] == (3, 8):
    from typing import Pattern
else:
    from re import Pattern


@validate_arguments
def build_prefixed_regex_pattern(
    *,
    prefix: str = Field(..., min_length=1),
    key_term: str = Field(..., min_length=1),
) -> Pattern[str]:
    """
    A regular expression builder that returns a compiled
    regular expression that matches a string if:
    - An escaped `prefix` string (may have whitespaces before or after) is followed by
    - The escaped `key_term` to capture with the group name `key`
    - Followed only by whitespace characters and/or the end of the line

    It will also match all the following newlines.
    """
    prefix = re.escape(prefix)
    key_term = re.escape(key_term)
    pattern = re.compile(
        rf"""
        ^ # match start of each line
        \s*{prefix}\s* # allow whitespace before and after prefix
        (?P<key>{key_term}) # term to capture
        \s* # allow any number of whitespaces after command
        $ # match end of each line (excludes \n in MULTILINE)
        [\r\n]* # Capture current and all following empty newlines
        """,
        re.VERBOSE | re.MULTILINE,
    )
    return pattern


@validate_arguments
def build_prefixed_regex_pattern_with_value(
    *,
    prefix: str = Field(..., min_length=1),
    key_term: str = Field(..., min_length=1),
    delimiter: str = Field("=", min_length=1),
    expand_key_term: bool = False,
) -> Pattern[str]:
    """
    A regular expression builder that returns a compiled
    regular expression that matches a string with:
    - The (escaped) `prefix` string (may have whitespaces before or after)
    - The (escaped) `key_term` to capture with the group name `key`
        - If `expand_key_term` is set, the `key_term` will be seen as a *key-prefix*
            and the matched key will be lazily expanded.
    - Followed by an (escaped) `delimiter` (may have whitespaces before or after)
    - and captures the following line until the end of the line with the group name `value`
    """
    prefix = re.escape(prefix)
    key_term = re.escape(key_term)
    delimiter = re.escape(delimiter)
    key_expansion_suffix = ".*?" if expand_key_term else ""
    pattern = re.compile(
        rf"""
        ^ # match start of each line
        \s*{prefix}\s* # allow whitespace before and after prefix
        (?P<key>{key_term}{key_expansion_suffix}) # term to capture
        \s*{delimiter}\s* # allow whitespace before and after delimiter
        (?P<value>[^\n\r]+)
        $ # match end of each line (excludes \n in MULTILINE)
        [\r\n]* # Capture current and all following empty newlines
        """,
        re.VERBOSE | re.MULTILINE,
    )
    return pattern



from typing import List, Union

import nbformat
import pytest

from common_nb_preprocessors.myst_nb import (
    MystNBCellConf,
    MystNBCellTags,
    myst_nb_metadata_injector,
)


@pytest.mark.parametrize("StrEnum", [MystNBCellConf, MystNBCellTags])
def test_cell_conf_enum_str(StrEnum):
    manual_value_str = {str(e.value) for e in StrEnum}
    implicit_value_str = {str(e) for e in StrEnum}
    assert manual_value_str == implicit_value_str


@pytest.mark.parametrize("prefix", ["/", "#", "###"])
@pytest.mark.parametrize("remove_line", [True, False])
@pytest.mark.parametrize("delimiter", ["|", "___"])
def test_myst_nb_metadata_injector(prefix: str, remove_line: bool, delimiter: str):
    tag_inp = tag_value = MystNBCellTags.remove_input
    cell_conf_inp = MystNBCellConf.remove_code_source
    cell_conf_value_inp = "true"
    cell_conf_value_output = True  # parsed as yaml
    nb = nbformat.v4.new_notebook()
    src = f"""\
        {prefix}{tag_inp}
        code line
        {prefix}{cell_conf_inp}{delimiter}{cell_conf_value_inp}
        other code line\
    """
    nb.cells.append(nbformat.v4.new_code_cell(src))
    new_nb = myst_nb_metadata_injector(
        nbformat.writes(nb), prefix=prefix, remove_line=remove_line, delimiter=delimiter
    )
    assert new_nb.cells[0].hasattr("metadata")
    assert new_nb.cells[0].metadata.hasattr("tags")
    assert new_nb.cells[0].metadata.tags == [tag_value]
    assert new_nb.cells[0].metadata.hasattr("mystnb")
    assert new_nb.cells[0].metadata.mystnb == {cell_conf_inp: cell_conf_value_output}
    len_source_lines = len(new_nb.cells[0].source.splitlines())
    if remove_line:
        assert len_source_lines == 2
    else:
        assert len_source_lines == 4


# test extra tags configurator
@pytest.mark.parametrize(
    "inp_tags,extra_tags",
    [
        (["tag_inp1"], None),
        (["tag_inp1", "tag_inp2"], None),
        (["tag_inp1"], ["tag_inp1"]),
        (["tag-inp1", "tag-inp2"], ["tag-inp1", "tag-inp2"]),
    ],
)
def test_myst_nb_metadata_injector_extra_tags(
    inp_tags: List[str], extra_tags: Union[List[str], None]
):
    # assumption: extra_tags always matches inp for all if not None
    nb = nbformat.v4.new_notebook()
    src_tmpl = """\
        # {inp_tag}
        code line
    """
    src_txt = "\n".join(src_tmpl.format(inp_tag=inp_tag) for inp_tag in inp_tags)
    nb.cells.append(nbformat.v4.new_code_cell(src_txt))
    new_nb = myst_nb_metadata_injector(
        nbformat.writes(nb), prefix="#", remove_line=True, extra_tags=extra_tags
    )
    if extra_tags is not None:
        assert new_nb.cells[0].metadata.hasattr("tags")
        assert set(new_nb.cells[0].metadata.tags) == set(extra_tags)
    else:
        assert not new_nb.cells[0].metadata.hasattr("tags")


def test_myst_nb_metadata_injector_syntax_sugar():
    nb = nbformat.v4.new_notebook()
    src = f"""\
        # figure.caption = I am a string
        # figure.caption_before = true
        other code line\
    """
    nb.cells.append(nbformat.v4.new_code_cell(src))
    new_nb = myst_nb_metadata_injector(
        nbformat.writes(nb),
        prefix="#",
        delimiter="=",
        remove_line=True,
    )
    assert new_nb.cells[0].hasattr("metadata")
    assert new_nb.cells[0].metadata.hasattr("mystnb")
    assert new_nb.cells[0].metadata.mystnb == {
        "figure": {"caption": "I am a string", "caption_before": True}
    }
    len_source_lines = len(new_nb.cells[0].source.splitlines())
    assert len_source_lines == 1



from nbformat.v4 import new_code_cell, new_notebook

from common_nb_preprocessors.keep import TagsKeepPreproccesor


def test_keep():
    nb = new_notebook()
    valid_data = "# keep code cell"
    keep_tag = "keep me"

    nb.cells.append(new_code_cell(valid_data, metadata={"tags": [keep_tag]}))
    nb.cells.append(new_code_cell("# not this one", metadata={"tags": ["empty"]}))

    nb, _ = TagsKeepPreproccesor(tags=[keep_tag]).preprocess(nb, None)
    assert len(nb.cells) == 1
    assert nb.cells[0]["source"] == valid_data


def test_keep2():
    valid_data = "# keep code cell"
    keep_tag = "keep me"
    nb = new_notebook()
    nb.cells.append(new_code_cell(valid_data, metadata={"tags": [keep_tag]}))
    nb.cells.append(new_code_cell("# not this one either"))

    nb, _ = TagsKeepPreproccesor(tags=[keep_tag]).preprocess(nb, None)
    assert nb.cells[0]["source"] == valid_data
    assert len(nb.cells) == 1



from collections.abc import Collection, MutableMapping
from functools import singledispatch
from typing import Any, Dict, List, Union


class NestedDictUpdateError(Exception):
    """Exception raised for errors while setting a nested dict value."""


class NestedDictTypeChangeError(NestedDictUpdateError):
    """Exception raised if setting of nested dict would change the type."""


class NestedDictCollectionOverwriteError(NestedDictUpdateError):
    """Exception raised if setting value of nested dict would overwrite collection."""


class NestedDictInvalidDictError(NestedDictUpdateError):
    """Exception raised if a dictionary is expected but other type is found during traversal."""


@singledispatch
def _nested_dict_updater_helper(d: object, keys: List[str], _value) -> None:
    raise NestedDictInvalidDictError(
        f"Expected a dictionary! Won't overwrite entry {d} to set nested keys {keys}"
    )


@_nested_dict_updater_helper.register
def _(d: MutableMapping, keys: List[str], value) -> None:
    if len(keys) == 1:
        final_key = keys[0]
        # should also check for other 'nested' types => always fail
        # also check if Type changes, if it does, also raise an Exception
        if final_key in d.keys():
            if not isinstance(d[final_key], type(value)):
                raise NestedDictTypeChangeError(
                    f"Will not overwrite value: {value} of type {type(value)} with {d[final_key]} of type: {type(d[final_key])}"
                )
            if isinstance(d[final_key], Collection) and not isinstance(
                d[final_key], str
            ):
                raise NestedDictCollectionOverwriteError(
                    f"Setting the value of entry '{final_key}' with value {value} would overwrite collection: {d[final_key]}"
                    + "\n"
                    + "Overwriting of collections is not allowed, even if the same collection type is used!"
                )
        d[keys[0]] = value
        return
    d = d.setdefault(keys[0], {})
    _nested_dict_updater_helper(d, keys[1:], value)


def nested_dict_updater(d: Union[Dict, Any], keys: List[str], value: Any) -> None:
    """**Inplace** nested dictionary updater."""
    try:
        _nested_dict_updater_helper(d, keys, value)
    except NestedDictUpdateError as e:
        raise NestedDictUpdateError(
            f"Error while setting nested dict {d} with key-nesting {keys} to value {value}"
        ) from e



from enum import Enum
from typing import List, Optional

import nbformat
from pydantic import validate_arguments

from .metadata_injector import (
    MetaDataListInjectorPreprocessor,
    MetaDataMapInjectorPreprocessor,
)

__all__ = [
    "MystNBCellTags",
    "MystNBCellConf",
    "myst_nb_metadata_injector",
]


class MystNBCellTags(str, Enum):
    """
    All different Myst-NB cell tag configuration options.
    """

    hide_input = "hide-input"
    """
    Hide the source code of the cell behind a collapsible button.
    See :ref:`sec/hide-input` for a visual example.
    """
    hide_output = "hide-output"
    """
    Hide the output of the cell behind a collapsible button.
    See :ref:`sec/hide-output` for a visual example.
    """
    hide_cell = "hide-cell"
    """
    Hide the input as well as the output of the cell behind a collapsible button.
    See :ref:`sec/hide-cell` for a visual example.
    """
    remove_input = "remove-input"
    """
    Removes the source code of the cell.
    See :ref:`sec/remove-input` for a visual example.
    """
    remove_stderr = "remove-stderr"
    """
    Only remove the standard error output of the code cell.
    See :ref:`sec/remove-stderr` for a visual example.
    """
    remove_output = "remove-output"
    """
    Remove the output of the code cell.
    See :ref:`sec/remove-output` for a visual example.
    """
    remove_cell = "remove-cell"
    """
    Remove the input as well as the output of the code cell.
    See :ref:`sec/remove-cell` for a visual example.
    """
    raises_exception = "raises-exception"
    """
    Continue execution normally if code cell raises exception.
    See :ref:`sec/raises-exception` for a visual example.
    """
    skip_execution = "skip-execution"
    """
    Do not execute code cell.
    See :ref:`sec/skip-execution` for a visual example.
    """

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _validate(cls) -> None:
        """
        Internal use only!
        Requires the development dependency pandas with
        the html-dependencies.
        Will parse the values from the website and check them
        with the values from the current file.
        If they aren't identical, a RuntimeError will be raised.
        """
        from ._constant_builder import _read_unique_myst_nb_table

        _myst_nb_cell_tags = _read_unique_myst_nb_table(match="remove-cell")
        cell_tags = _myst_nb_cell_tags["Tag"].tolist()
        _nbclient_cell_tags = _read_unique_myst_nb_table(match="skip-execution")
        cell_tags.extend(_nbclient_cell_tags["Tag"].tolist())
        # if some aren't documented yet, which has happened quite often
        if set(cell_tags) - set(cls) != set():
            raise RuntimeError(f"{cls} out-of-date!")


class MystNBCellConf(str, Enum):
    """
    All different Myst-NB cell metadata configuration options.
    These keys *must* be paired with a value.
    """

    remove_code_source = "remove_code_source"
    """
    Remove the code cell.
    Should be paired with a boolean: `true`/`false`.
    See :ref:`sec/remove_code_source` for a visual example.
    """

    remove_code_outputs = "remove_code_outputs"
    """
    Remove the output of the cell.
    Should be paired with a boolean: `true`/`false`.
    See :ref:`sec/remove_code_outputs` for a visual example.
    """

    number_source_lines = "number_source_lines"
    """
    Add source line numbering to the current cell.
    Should be paired with a boolean: `true`/`false`.
    See :ref:`sec/number_source_lines` for a visual example.
    """

    output_stderr = "output_stderr"
    """
    Define the behavior for standard error output.
    Should be paired with one of the following strings:

    - `show`: Show the standard error as a separate output
    - `remove`: Remove the standard error output
    - `remove-warn`: Remove the standard error output but log a warning to Sphinx
    - `warn`/`error`/`severe`: Log the error with a specific logging level to Sphinx

    See :ref:`sec/output_stderr` for a visual example.
    """
    code_prompt_show = "code_prompt_show"
    """
    Define the prompt *text* that is displayed next to the
    *expand/show* dialog. The text is displayed to inform the user
    that the dialog can be expanded.

    The optional `{type}` placeholder will be replaced with:
    `content`, `source`, or `outputs`, depending on the `hide` tag.

    See :ref:`sec/code_prompt_show` for a visual example.
    """
    code_prompt_hide = "code_prompt_hide"
    """
    Define the prompt *text* that is displayed next to the
    *hide* dialog. The text is displayed to inform the user
    that the dialog can be *collapsed/minimized*.

    The optional `{type}` placeholder will be replaced with:
    `content`, `source`, or `outputs`, depending on the `hide` tag.

    See :ref:`sec/code_prompt_hide` for a visual example.
    """

    merge_streams = "merge_streams"
    """
    Merge standard output and standard error into a single output cell.
    Should be paired with a boolean: `true`/`false`.

    See :ref:`sec/merge_streams` for a visual example.
    """

    text_lexer = "text_lexer"
    """
    Pygments lexer applied to standard out, standard error and text/plain outputs.
    Should be paired with a Pygments lexer string.

    See :ref:`sec/text_lexer` for a visual example.
    """

    error_lexer = "error_lexer"
    """
    Pygments lexer applied to the error/traceback outputs.
    Should be paired with a Pygments lexer string.

    See :ref:`sec/error_lexer` for a visual example.
    """

    markdown_format = "markdown_format"
    """
    The format to use for text/markdown rendering.
    Should be paired with one of the following strings:

    - `commonmark` (Default)
    - `gfm`
    - `myst`

    See :ref:`sec/markdown_format` for a visual example.
    """

    image = "image"
    """
    Options for image outputs.


    See :ref:`sec/image` for a visual example.
    """

    figure = "figure"
    """
    Options for figure outputs.
    To see a list of all possible configuration values,
    take a look at the `official figure documentation <https://myst-nb.readthedocs.io/en/latest/render/format_code_cells.html#images-and-figures>`_.

    See :ref:`sec/figure` for a visual example.
    """

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _validate(cls) -> None:
        """
        Internal use only!
        Requires the development dependency pandas with
        the html-dependencies.
        Will parse the values from the website and check them
        with the values from the current file.
        If they aren't identical, a RuntimeError will be raised.
        """
        from ._constant_builder import _read_unique_myst_nb_table

        cell_tags = _read_unique_myst_nb_table("^markdown_format$")["Name"].tolist()
        if set(cell_tags) != set(cls):
            raise RuntimeError(f"{cls} out-of-date!")


@validate_arguments()
def myst_nb_metadata_injector(
    file_content: str,
    /,
    prefix: str = "#",
    remove_line: bool = True,
    delimiter: str = "=",
    extra_tags: Optional[List[str]] = None,
) -> nbformat.NotebookNode:
    """
    The preprocessor will inject all the MyST-NB specific tags into the
    metadata `tags` and the cell level configuration to the `mystnb` field.
    For an in-depth explanation of the different values, see the
    `cell-config` and the `tag` section of the MyST-NB documentation.

    :param file_content: contents of an `ipynb` file
    :param prefix: Comment symbol that precedes the keys. Defaults to `#`.
    :param remove_line: Set if the metadata comment lines should be removed after injection. Defaults to `True`.
    :param extra_tags: Additional custom `tags` list for further customization.

    Internally it calls `MetaDataListInjectorPrepreprocessor` and
    `MetaDataMapInjectorPreprocessor`.
    Refer to those classes for more details.
    """
    nb = nbformat.reads(file_content, as_version=4)
    metadata_tags = set(s.value for s in MystNBCellTags)
    if extra_tags is not None:
        metadata_tags = metadata_tags | set(extra_tags)

    # could be done in one preprocess step
    nb, _ = MetaDataListInjectorPreprocessor(
        metadata_group="tags",
        strings=list(metadata_tags),
        prefix=prefix,
        remove_line=remove_line,
    ).preprocess(nb, None)
    nb, _ = MetaDataMapInjectorPreprocessor(
        metadata_group="mystnb",
        keys=list(MystNBCellConf),
        prefix=prefix,
        remove_line=remove_line,
        delimiter=delimiter,
        value_to_yaml=True,
        allow_nested_keys=True,
    ).preprocess(nb, None)
    return nb


# def jupyter_book_metadata_injector(
#     file_content: str, prefix: str = "#", remove_line: bool = True
# ):
#     """
#     The preprocessor will inject all the jupyter-book specific tags into the
#     metadata tags group of the code cells.
#     For extra information about the jupyter-book specific tags, see


#     By default, the `prefix` or `comment` symbol is assumed to be `#`
#     and the line containing the magic comment is removed.

#     Args:
#         file_content (str): contents of an `ipynb` file
#         prefix (str, optional): Comment symbol that preceeds the jupyter-book keys. Defaults to "#".
#         remove_line (bool, optional): Set if the metadata comment lines should be removed after injection. Defaults to True.
#     """
#     inp_nb = nbformat.reads(file_content, as_version=4)
#     nb, _ = MetaDataListInjectorPreprocessor(
#         strings=JUPYTER_BOOK_CODE_TAGS, prefix=prefix, remove_line=remove_line
#     ).preprocess(inp_nb, None)
#     return nb



from typing import Dict, Tuple, Union

import traitlets
from nbconvert.preprocessors import Preprocessor
from nbformat import NotebookNode

__all__ = ["TagsKeepPreprocessor"]


class TagsKeepPreproccesor(Preprocessor):
    """
    Only keep *code* cells that have the given metadata
    """

    tags = traitlets.List(traitlets.Unicode()).tag(config=True)

    def preprocess(
        self, nb: NotebookNode, resources: Union[None, Dict]
    ) -> Tuple[NotebookNode, Union[Dict, None]]:
        nb.cells = [
            cell
            for cell in nb.cells
            if set(self.tags).intersection(
                set(cell.get("metadata", {}).get("tags", []))
            )
            or cell["cell_type"] != "code"
        ]
        return nb, resources



import hypothesis.strategies as st
import pytest
from hypothesis import given, settings
from pydantic import ValidationError

from common_nb_preprocessors._patterns import (
    build_prefixed_regex_pattern,
    build_prefixed_regex_pattern_with_value,
)

allowed_pattern_st = st.text(
    alphabet=st.characters(
        blacklist_categories=("C", "Z"), blacklist_characters=("\n", "\r")
    ),
    min_size=1,
)
spacer_str = st.characters(whitelist_categories=("Z",), blacklist_characters=("\n"))


@given(
    spacer_str,
    allowed_pattern_st,
    spacer_str,
    allowed_pattern_st,
)
@settings(max_examples=250)
def test_metadata_list_injector_parsing(prefix_spacer, prefix, spacer, key_term):
    text = f"{prefix_spacer}{prefix}{spacer}{key_term}"
    pattern = build_prefixed_regex_pattern(prefix=prefix, key_term=key_term)
    assert pattern.search(text).group("key") == key_term


def test_pattern_builder():
    pattern = build_prefixed_regex_pattern(prefix="#", key_term="hide")
    assert "hide" == pattern.search("#hide").group("key")
    assert "hide" == pattern.search("# hide").group("key")
    assert "hide" == pattern.search("#  hide").group("key")


@pytest.mark.parametrize("invalid_kwargs", [{"prefix": ""}, {"key_term": ""}])
def test_prefixed_regex_pattern_invalid_inp(invalid_kwargs):
    with pytest.raises(ValidationError):
        build_prefixed_regex_pattern(**invalid_kwargs)


@given(
    allowed_pattern_st,
    spacer_str,
    allowed_pattern_st,
    allowed_pattern_st,
    allowed_pattern_st,
)
@settings(max_examples=250)
def test_metadata_map_injector_parsing(prefix, spacer, key_term, delimiter, value_term):
    text = f"{prefix}{spacer}{key_term}{delimiter}{value_term}"
    pattern = build_prefixed_regex_pattern_with_value(
        prefix=prefix, key_term=key_term, delimiter=delimiter
    )
    assert pattern.search(text).group("key") == key_term
    assert pattern.search(text).group("value") == value_term


@pytest.mark.parametrize(
    "text",
    [
        "# default_exp=export_name",
        "# default_exp= export_name",
        "# default_exp =export_name",
        "# default_exp\t= export_name",
        "#default_exp=export_name",
        "#\tdefault_exp=export_name",
    ],
)
def test_pattern_with_value_builder_whitespace_handling(text):
    pattern2 = build_prefixed_regex_pattern_with_value(
        prefix="#",
        key_term="default_exp",
        delimiter="=",
    )
    assert "default_exp" == pattern2.search(text).group("key")
    assert "export_name" == pattern2.search(text).group("value")


@pytest.mark.parametrize(
    "delimiter,text",
    [
        ("=", "# hide=true"),
        ("=", "# hide = true"),
        ("=", "# hide\t=\ttrue"),
        ("/ ", "# hide/ true"),
        (" ", "# hide true"),
        ("==", "# hide==true"),
        ("_-_", "# hide_-_true"),
        ("\\", r"# hide \ true"),
    ],
)
def test_pattern_with_value_delimiter_handling(delimiter, text):
    pattern = build_prefixed_regex_pattern_with_value(
        prefix="#", key_term="hide", delimiter=delimiter
    )
    assert "hide" == pattern.search(text).group("key")
    assert "true" == pattern.search(text).group("value")


# If I am bored, I could rewrite these with hypothesis
@pytest.mark.parametrize("delimiter", ["="])
# hide will always be key_term!
@pytest.mark.parametrize(
    "inp_key_text,matched_key",
    [
        ("hide", "hide"),
        ("hide ", "hide"),
        ("hide.", "hide."),
        ("hide. ", "hide."),
        ("hide.two", "hide.two"),
        ("hide.two ", "hide.two"),
    ],
)
def test_pattern_with_value_key_expansion(delimiter, inp_key_text, matched_key):
    prefix = "#"
    key = "hide"
    value = "true"
    text = f"{prefix} {inp_key_text}{delimiter}{value}"
    pattern = build_prefixed_regex_pattern_with_value(
        prefix=prefix, key_term=key, delimiter=delimiter, expand_key_term=True
    )
    assert matched_key == pattern.search(text).group("key")
    assert "true" == pattern.search(text).group("value")


@pytest.mark.parametrize(
    "invalid_kwargs", [{"prefix": ""}, {"key_term": ""}, {"delimiter": ""}]
)
def test_pattern_with_value_invalid_inp(invalid_kwargs):
    with pytest.raises(ValidationError):
        build_prefixed_regex_pattern_with_value(**invalid_kwargs)


def test_pattern_with_value_newline_matching():
    pattern = build_prefixed_regex_pattern_with_value(
        prefix="#", key_term="hide", delimiter="="
    )
    assert "import lib" == pattern.sub("", "# hide=true\nimport lib")
    assert "import lib" == pattern.sub("", "# hide=true\n\nimport lib")



import nbformat
import pytest

from common_nb_preprocessors._nested_dict_updater import NestedDictUpdateError
from common_nb_preprocessors.metadata_injector import (
    MetaDataListInjectorPreprocessor,
    MetaDataMapInjectorPreprocessor,
)


def test_metadata_list_injector():
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell("# hide\nimport os"))
    nb.cells.append(nbformat.v4.new_markdown_cell("# This is the actual title"))
    nb, _ = MetaDataListInjectorPreprocessor(strings=("hide",), prefix="#").preprocess(
        nb, None
    )
    assert nb.cells[0]["source"] == "import os"
    assert nb.cells[0]["metadata"]["tags"][0] == "hide"
    # check idempotency
    nb, _ = MetaDataListInjectorPreprocessor(strings=("hide",), prefix="#").preprocess(
        nb, None
    )
    assert nb.cells[0]["metadata"]["tags"][0] == "hide"


@pytest.mark.parametrize(
    "cell_source_inp,cell_source_output,remove_line",
    [
        ("# hide\nimport os", "import os", True),
        ("# hide\nimport os", "# hide\nimport os", False),
    ],
)
def test_metadata_list_injector_remove_line(
    cell_source_inp, cell_source_output, remove_line
):
    strings = ["hide"]
    prefix = "#"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(cell_source_inp))
    nb, _ = MetaDataListInjectorPreprocessor(
        strings=strings, prefix=prefix, remove_line=remove_line
    ).preprocess(nb, None)
    assert nb.cells[0]["source"] == cell_source_output
    # check that parsing works as expected
    assert nb.cells[0]["metadata"]["tags"][0] == "hide"


def test_metadata_list_injector_empty_metadata_group():
    with pytest.raises(ValueError, match="non-empty"):
        MetaDataListInjectorPreprocessor(strings=["a"], metadata_group="")


@pytest.mark.parametrize(
    "cell_source_inp,cell_source_output,remove_line",
    [
        ("# hide=true\nimport os", "import os", True),
        ("# hide=true\nimport os", "# hide=true\nimport os", False),
    ],
)
def test_metadata_map_injector_remove_line(
    cell_source_inp, cell_source_output, remove_line
):
    keys = ["hide"]
    prefix = "#"
    delimiter = "="
    metadata_group = "mystnb"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(cell_source_inp))
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=keys,
        prefix=prefix,
        remove_line=remove_line,
        delimiter=delimiter,
        metadata_group=metadata_group,
        allow_nested_keys=False,
    ).preprocess(nb, None)
    assert nb.cells[0]["source"] == cell_source_output
    # check that parsing works as expected
    assert nb.cells[0]["metadata"][metadata_group] == {"hide": "true"}


@pytest.mark.parametrize(
    "delimiter,cell_comment",
    [
        ("=", "# hide=true"),
        ("\\", r"# hide \ true"),
    ],
)
def test_metadata_map_injector_delimiter(delimiter, cell_comment):
    keys = ["hide"]
    prefix = "#"
    metadata_group = "mystnb"
    code_cell_text = f"{cell_comment}\nimport lib"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(code_cell_text))
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=keys,
        prefix=prefix,
        metadata_group=metadata_group,
        delimiter=delimiter,
        remove_line=True,
    ).preprocess(nb, None)
    injected_map = nb.cells[-1]["metadata"]["mystnb"]
    assert list(injected_map.keys()) == keys
    assert injected_map[keys[0]] == "true"


def test_metadata_map_injector_multiple_matches():
    code_cell_text = f"# hide = true\n# hide = false\nimport lib"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(code_cell_text))
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=["hide"],
        prefix="#",
        metadata_group="mystnb",
        delimiter="=",
        remove_line=True,
        allow_nested_keys=False,
    ).preprocess(nb, None)
    injected_map = nb.cells[0]["metadata"]["mystnb"]
    assert list(injected_map.keys()) == ["hide"]
    assert len(nb.cells[0]["source"].splitlines()) == 1
    assert injected_map["hide"] == "false"


@pytest.mark.parametrize(
    "inp_comment, output_val",
    [
        ("hide = true", True),
        ("hide = false", False),
        ("hide = I am a string", "I am a string"),
        ("hide = 42", 42),
    ],
)
def test_metadata_map_injector_multiple_to_yaml(inp_comment, output_val):
    code_cell_text = f"# {inp_comment}\nimport lib"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(code_cell_text))
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=["hide"],
        prefix="#",
        metadata_group="mystnb",
        delimiter="=",
        remove_line=True,
        allow_nested_keys=False,
        value_to_yaml=True,
    ).preprocess(nb, None)
    injected_map = nb.cells[0]["metadata"]["mystnb"]
    assert list(injected_map.keys()) == ["hide"]
    assert injected_map["hide"] == output_val


def test_metadata_map_injector_multiple_nested_access():
    code_cell_text = f"# hide.nested = true\n# hide.other = false\nimport lib"
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(code_cell_text))
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=["hide"],
        prefix="#",
        metadata_group="mystnb",
        delimiter="=",
        remove_line=True,
        allow_nested_keys=True,
    ).preprocess(nb, None)
    injected_map = nb.cells[0]["metadata"]["mystnb"]
    assert list(injected_map.keys()) == ["hide"]
    assert len(nb.cells[0]["source"].splitlines()) == 1
    assert injected_map["hide"] == {"nested": "true", "other": "false"}


def test_metadata_map_injector_empty_metadata_group():
    nb = nbformat.v4.new_notebook()
    with pytest.raises(ValueError, match="non-empty"):
        nb, _ = MetaDataMapInjectorPreprocessor(
            keys=["a"], metadata_group=""
        ).preprocess(nb, None)


def test_metadata_list_markdown():
    nb = nbformat.v4.new_notebook()
    source = "# a"
    nb.cells.append(
        nbformat.v4.new_markdown_cell(source),
    )
    nb.cells.append(
        nbformat.v4.new_code_cell(source),
    )
    nb, _ = MetaDataListInjectorPreprocessor(strings=["a"], prefix="#").preprocess(
        nb, None
    )
    assert nb.cells[0].source == source
    assert nb.cells[1].source == ""


def test_metadata_map_markdown():
    nb = nbformat.v4.new_notebook()
    source = "# a = true"
    nb.cells.append(
        nbformat.v4.new_markdown_cell(source),
    )
    nb.cells.append(
        nbformat.v4.new_code_cell(source),
    )
    nb, _ = MetaDataMapInjectorPreprocessor(
        keys=["a"],
        prefix="#",
        delimiter="=",
        metadata_group="mystnb",
    ).preprocess(nb, None)
    assert nb.cells[0].source == source
    assert nb.cells[1].source == ""


def test_metadata_list_type_clash():
    nb = nbformat.v4.new_notebook()
    nb.cells.append(
        nbformat.v4.new_code_cell("# a", metadata={"other": "single-value"})
    )
    with pytest.raises(RuntimeError, match="type"):
        nb, _ = MetaDataListInjectorPreprocessor(
            strings=["a"], metadata_group="other"
        ).preprocess(nb, None)


def test_metadata_map_type_clash():
    nb = nbformat.v4.new_notebook()
    nb.cells.append(
        nbformat.v4.new_code_cell("# a=false", metadata={"mystnb": "single-value"})
    )
    with pytest.raises(NestedDictUpdateError):
        nb, _ = MetaDataMapInjectorPreprocessor(
            keys=["a"], metadata_group="mystnb"
        ).preprocess(nb, None)



import sys

import pytest
from pytest import MonkeyPatch

import common_nb_preprocessors._constant_builder
from common_nb_preprocessors.myst_nb import MystNBCellConf, MystNBCellTags


# FUTURE: Figure out how to cache the pandas parsing
def test_mystnb_cell_tags_constant():
    MystNBCellTags._validate()


def test_mystnb_cell_conf_constant():
    MystNBCellConf._validate()


def test_unsynced_mystnb_cell_tags_constant(monkeypatch):
    def a(_):
        yield

    # TODO: understand why this is the case!
    monkeypatch.setattr(MystNBCellTags.__class__, "__iter__", a)
    with pytest.raises(RuntimeError):
        MystNBCellTags._validate()


def test_unsynced_mystnb_cell_conf_constant(monkeypatch):
    def a(_):
        yield

    # TODO: understand why this is the case!
    monkeypatch.setattr(MystNBCellConf.__class__, "__iter__", a)
    with pytest.raises(RuntimeError):
        MystNBCellConf._validate()


@pytest.mark.parametrize("without_lib", ["pandas", "bs4", "html5lib"])
def test_import_error(without_lib: str, monkeypatch: MonkeyPatch):
    monkeypatch.setitem(sys.modules, without_lib, None)

    # reload as otherwise monkey-patching was not sufficient
    # when running all tests
    import importlib

    with pytest.raises(ImportError, match="requires"):
        importlib.reload(common_nb_preprocessors._constant_builder)


def test_multiple_matching_tables(monkeypatch: MonkeyPatch):
    def return_two_elem_list(*args, **kwargs):
        return [1, 2]

    import pandas

    monkeypatch.setattr(pandas, "read_html", return_two_elem_list)
    with pytest.raises(ValueError, match="unique table"):
        common_nb_preprocessors._constant_builder._read_unique_myst_nb_table(
            "IamNotOnTheMySTPageAndWillNotFindAMatch"
        )



import pytest

from common_nb_preprocessors._nested_dict_updater import (
    NestedDictCollectionOverwriteError,
    NestedDictInvalidDictError,
    NestedDictTypeChangeError,
    NestedDictUpdateError,
    nested_dict_updater,
)


@pytest.mark.parametrize(
    "inp_dict,keys,value,out_dict",
    [
        ({}, ["a"], 1, {"a": 1}),
        ({}, ["a"], "1", {"a": "1"}),
        ({}, ["a", "b"], 1, {"a": {"b": 1}}),
        ({}, ["a", "b", "c"], 1, {"a": {"b": {"c": 1}}}),
        ({"a": 1}, ["a"], 2, {"a": 2}),
        ({"a": {"b": {"c": "z"}}}, ["a", "b", "c"], "1", {"a": {"b": {"c": "1"}}}),
        ({"a": 1}, ["b"], 2, {"a": 1, "b": 2}),
        (
            {"a": {"b": 1, "c": 2}, "d": 3},
            ["a", "b"],
            9,
            {"a": {"b": 9, "c": 2}, "d": 3},
        ),
    ],
)
def testnested_dict_updater(inp_dict, keys, value, out_dict):
    # works in-place!
    nested_dict_updater(inp_dict, keys, value)
    assert inp_dict == out_dict


@pytest.mark.parametrize(
    "inp_dict,keys",
    [
        (["not_dict"], ["a"]),
        ({"a": 1}, ["a", "b"]),
        ({"a": {"b": 1}}, ["a", "b", "c"]),
    ],
)
def testnested_dict_updater_invalid_nesting(inp_dict, keys):
    with pytest.raises(NestedDictUpdateError) as exc_info:
        nested_dict_updater(inp_dict.copy(), keys, {"b": 1})
    assert type(exc_info.value.__cause__) == NestedDictInvalidDictError


@pytest.mark.parametrize(
    "inp_dict,keys,value",
    [
        ({"a": {"b": 1}}, ["a"], 1),
        ({"a": {"b": 1}}, ["a", "b"], "1"),
        ({"a": {"b": 1}}, ["a", "b"], [1]),
    ],
)
def testnested_dict_updater_invalid_type_setting(inp_dict, keys, value):
    with pytest.raises(NestedDictUpdateError) as exc_info:
        nested_dict_updater(inp_dict.copy(), keys, value)
    assert type(exc_info.value.__cause__) is NestedDictTypeChangeError


@pytest.mark.parametrize(
    "inp_dict,keys,value",
    [
        ({"a": {"b": 1}}, ["a"], {"b": 2, "c": 1}),
        ({"a": {"b": 1}}, ["a"], {"b": 1}),
        ({"a": [1]}, ["a"], [2]),
        ({"a": {1}}, ["a"], {2}),
    ],
)
def testnested_dict_updater_invalid_collection_setting(inp_dict, keys, value):
    with pytest.raises(NestedDictUpdateError) as exc_info:
        nested_dict_updater(inp_dict.copy(), keys, value)
    assert type(exc_info.value.__cause__) is NestedDictCollectionOverwriteError



try:
    # TODO: check set html5lib as dependency
    import bs4  # type: ignore
    import html5lib  # type: ignore
    import lxml  # type: ignore
    import pandas
except ImportError:
    raise ImportError(
        """\
        This module is only used during development and requires
        pandas + the `read_html` dependencies (bs4, html5lib, and lxml).
        """
    )


def _read_unique_pd_table(url: str, match: str) -> pandas.DataFrame:
    dfs = pandas.read_html(url, match=match)
    if len(dfs) > 1:
        raise ValueError("Match does not provide a unique table for url: ", url)
    return dfs[0]


def _read_unique_myst_nb_table(match: str) -> pandas.DataFrame:
    return _read_unique_pd_table(
        url="https://myst-nb.readthedocs.io/en/latest/configuration.html", match=match
    )


# def _read_unique_jupyter_book_table(match: str) -> pandas.DataFrame:
#     return _read_unique_pd_table(
#         url="https://jupyterbook.org/en/latest/reference/cheatsheet.html#myst-cheatsheet-code-cell-tags",
#         match=match,
#     )


# def jupyter_book_cell_tags() -> List[str]:
#     """
#     Return Jupyter-Book tags.
#     """
#     df = _read_unique_jupyter_book_table("hide-input")
#     tag_col = df["Tag option"].astype(str).str.strip("'\"")
#     return tag_col.tolist()



from typing import Dict, List, Tuple, Union

import traitlets
import yaml
from nbconvert.preprocessors import Preprocessor
from nbformat import NotebookNode

from ._nested_dict_updater import nested_dict_updater
from ._patterns import (
    build_prefixed_regex_pattern,
    build_prefixed_regex_pattern_with_value,
)

__all__ = [
    "MetaDataListInjectorPreprocessor",
    "MetaDataMapInjectorPreprocessor",
]


class MetaDataListInjectorPreprocessor(Preprocessor):
    """
    Parse all *code* cells and append the matched magic comments with the
    `prefix` to the `metadata_group` list.
    These `strings` must be on their own line and only contain the `prefix`,
    a `string` from `strings` (i.e., the magic comment) and whitespace characters.
    """

    metadata_group: str = traitlets.Unicode(default_value="tags").tag(config=True)
    """Metadata group to which the matched magic comment will be appended to if
    it doesn't already exist. Default is `tags`."""
    strings: List[str] = traitlets.List(traitlets.Unicode(), minlen=1).tag(config=True)
    """List of strings (magic comments) that define the text that will be matched and
    injected into the selected metadata group."""
    prefix: str = traitlets.Unicode(default_value="#").tag(config=True)
    """The prefix that indicates the possible start of a magic comment line.
    Should be comment character of the language. By default `#`."""
    remove_line: bool = traitlets.Bool(default_value=True).tag(config=True)
    """By default remove the matching line in the code-cell."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.metadata_group == "":
            raise ValueError("metadata_group myst be non-empty string!")

    def _write_tag(self, tag: str, cell: NotebookNode) -> NotebookNode:
        tags = cell.setdefault("metadata", {}).setdefault(self.metadata_group, [])
        if not isinstance(tags, list):
            raise RuntimeError(
                f"Trying to to set metadata list but entry {self.metadata_group} is of type: {type(tags)}",
                self.metadata_group,
            )
        if tag not in tags:
            tags.append(tag)
        cell["metadata"][self.metadata_group] = tags
        return cell

    def preprocess_cell(
        self, cell: NotebookNode, resource: Union[Dict, None], _index: int
    ) -> Tuple[NotebookNode, Union[Dict, None]]:
        """Inject metadata to code-cell if match is found"""
        if cell["cell_type"] == "markdown":
            return cell, resource
        for string in self.strings:
            pattern = build_prefixed_regex_pattern(prefix=self.prefix, key_term=string)
            matches = pattern.finditer(cell.source)
            for m in matches:
                tag = m.group("key")
                cell = self._write_tag(tag, cell)
            if self.remove_line:
                cell.source = pattern.sub("", cell.source)
        return cell, resource


class MetaDataMapInjectorPreprocessor(Preprocessor):
    """
    Parse all *code* cells and add the matched key-value pairs with the
    `prefix` to the `metadata_group` dictionary.
    The key-value pairs are generated by searching for each `key` of `keys` followed
    by `delimiter` and the value.
    """

    metadata_group: str = traitlets.Unicode().tag(config=True)
    """Metadata group into which the matched key-value pairs will be written."""
    keys: List[str] = traitlets.List(traitlets.Unicode()).tag(config=True)
    """List of keys that will be used as a key for the `metadata_group` dictionary entry and is followed by the `delimiter` and `value`."""
    prefix: str = traitlets.Unicode(default_value="#").tag(config=True)
    """The prefix that indicates the possible start of a magic comment line. Should be comment character of the language."""
    remove_line: bool = traitlets.Bool(default_value=True).tag(config=True)
    """By default remove the matching line in the code-cell."""
    delimiter: str = traitlets.Unicode(default_value="=").tag(config=True)
    """Delimiter that separates the key from the value."""
    value_to_yaml: bool = traitlets.Bool(default_value=False).tag(config=True)
    """Parse the value as yaml syntax before writing it as a dictionary. Default is `False`."""
    allow_nested_keys: bool = traitlets.Bool(default_value=False).tag(config=True)
    """Allow defining nested key access for nested setting of metadata group. Access next level with `.` (hardcoded)"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.metadata_group == "":
            raise ValueError("metadata_group myst be non-empty string!")

    def preprocess_cell(
        self, cell: NotebookNode, resource: Union[Dict, None], _index: int
    ) -> Tuple[NotebookNode, Union[Dict, None]]:
        """Inject metadata dict entry to code-cell if match is found"""
        if cell["cell_type"] == "markdown":
            return cell, resource
        for key_term in self.keys:
            pattern = build_prefixed_regex_pattern_with_value(
                prefix=self.prefix,
                key_term=key_term,
                delimiter=self.delimiter,
                expand_key_term=self.allow_nested_keys,
            )
            matches = pattern.finditer(cell.source)
            for m in matches:
                value = m.group("value")
                parsed_value = yaml.safe_load(value) if self.value_to_yaml else value
                key = m.group("key")  # Could be expanded key!
                keys = key.split(".") if self.allow_nested_keys else [key]
                entry = cell.setdefault("metadata", {}).setdefault(
                    self.metadata_group, {}
                )
                nested_dict_updater(entry, keys, parsed_value)
            if self.remove_line:
                cell.source = pattern.sub("", cell.source)
        return cell, resource


# class GlobalMetaDataInjectorPreprocessor(Preprocessor):
#     """
#     Parse all *code* cells and convert the matching `prefix` `key` `value`
#     lines to the global `metadata` field.

#     To clean up the output, the lines containing any `string` may be removed
#     by setting `remove_line=True` (default).

#     The provided list of `keys` will be used to access the *global* `metadata` field
#     and insert the value that is followed by the `key` in the code cell.
#     Note that the global metadata field will be overwritten if multiple cells define the
#     field's value.
#     """

#     keys = List(Unicode()).tag(config=True)
#     prefix = Unicode(default_value="#").tag(config=True)
#     delimiter = Unicode(default_value=r"=").tag(config=True)

#     def preprocess(self, nb, resources):
#         if len(self.keys) == 0:
#             return nb, resources

#         for cell in nb.cells:
#             if cell["cell_type"] == "markdown":
#                 continue
#             for key in self.keys:
#                 pattern = build_prefixed_regex_pattern_with_value(
#                     self.prefix, key, delimiter=self.delimiter
#                 )
#                 m = pattern.search(cell.source)
#                 if m is not None:
#                     value = m.group("value")
#                     nb.setdefault("metadata", {})
#                     nb["metadata"][key] = value
#         return nb, resources
